
% EKF   Extended Kalman Filter for nonlinear state estimation.
%
%   [xhat,yhat,P]=EKF(mod, data, x0, P0) implements
%   the prediction and update steps of the Extended Kalman Filter for a 
%   nonlinear system. The function estimates the current state and its 
%   covariance based on a nonlinear process model and nonlinear measurement
%   model.
%
%   Inputs:
%       mod.f  - Function handle for the nonlinear state transition model
%                x_{k+1} = f(x_k, u_k, w_k)
%       mod.h  - Function handle for the nonlinear measurement model
%                z_k = h(x_k, u_k) + v_k
%       mod.x0 - The intial state estimate (n x 1 vector)
%       mod.nx - The number of states.
%       mod.nw - The dimension of the process noise w.
%       mod.P0 - The inital state covariance matrix (nx x nx matrix)
%       mod.Q  - Process noise w_k covariance matrix (nw x nw matrix)
%       mod.R  - Measurement noise v_k covariance matrix (m x m matrix)
%       mod.fx - Function handle for the Jacobian of the state transition model
%                (nx x nx matrix evaluated at x_k and u)
%       mod.hx - Function handle for the Jacobian of the measurement model
%                (m x nx matrix evaluated at x_k and u)
%
%       data.u - The control input (optional, p x N vector)
%       data.z - The output measurement (m x N vector)
%
%   Outputs:
%       x_hat  - Updated state estimate (nx x N vector)
%       y_hat  - Updated output estimate (m x N vector)
%       P      - Updated state covariance matrix (cell array of length N where each cell contains an nx x nx matrix)
%
%   Description:
%       The Extended Kalman Filter is used for state estimation in 
%       nonlinear dynamic systems. It assumes that the process and 
%       measurement models are nonlinear, but can be approximated linearly 
%       using Jacobians.
%
%       The function performs two main steps:
%       1. **Prediction step**: Predict the state and covariance based on
%          the nonlinear process model.
%       2. **Update step**: Update the predicted state and covariance
%          using the current measurement and the nonlinear measurement model.
%
%   Example:
%       % Define nonlinear models
%
%       mod.nx = 2; % Number of states
%       mod.nw = 2; % Dimension of process noise w
%
%       mod.f = @(x,u,w) [x(1) + u(1) + w(1); x(2) + u(2) + w(2)];
%       mod.h = @(x,u) [sqrt(x(1)^2 + x(2)^2)];
%       mod.fx = @(x,u,w) [1 0; 0 1];  % Jacobian of df/dx
%       mod.fw = @(x,u,w) [1 0; 0 1];  % Jacobian of df/dw
%       mod.hx = @(x,u) [x(1)/sqrt(x(1)^2 + x(2)^2), x(2)/sqrt(x(1)^2 + x(2)^2)]; % Jacobian of dh/dx
%
%       % Initial conditions
%       x0 = [1; 0];
%       P0 = eye(2);
%
%       % Define covariance matrices
%       mod.R = 1;
%       mod.Q = [1 .5; .5 2];
%
%       % Define input and output data
%       data.u = u;
%       data.y = y;
%
%       % Perform EKF
%       [xhat,yhat,P]=EKF(mod, data, x0, P0);
%


function [xhat,yhat,P]=EKF(mod, data, x0, P0)

n = mod.nx;
m = size(data.y,1);
N = size(data.y,2);

if nargin < 3
    x0 = zeros(n,1);
end
if nargin < 4
    P0 = eye(n);
end
filt.xhat = x0;
filt.P    = P0;

xhat = zeros(n,N); xhat(:,1) = x0;
P = cell(1, N);
yhat = zeros(m,N);

for k=1:N-1
    % Measurement update
    [xhat(:,k), filt] = measUpdate(filt, mod, data.y(:,k), data.u(:,k));
    P{k} = filt.P;
    yhat(:,k) = mod.h(xhat(:,k),data.u(:,k));
    % Time update
    [xhat(:,k+1), filt] = timeUpdate(filt, mod, data.u(:,k));
end
[xhat(:,N), filt] = measUpdate(filt, mod, data.y(:,N), data.u(:,N));
P{N} = filt.P;
yhat(:,N) = mod.h(xhat(:,N),data.u(:,N));
end

function [xhat, filt] = measUpdate(filt, mod, y, u)

xhat = filt.xhat;
C = mod.hx(xhat,u);

% Measurement update
K=filt.P*C'*inv(C*filt.P*C'+mod.R);
filt.P=filt.P-K*C*filt.P;
xhat = xhat + K*(y-mod.h(xhat,u));
filt.xhat = xhat;
end


function [xhat, filt] = timeUpdate(filt,mod,u)

nw = mod.nw;
xhat = filt.xhat;
F = mod.fx(xhat,u,zeros(nw,1));
G = mod.fw(xhat,u,zeros(nw,1));

% Time update
xhat = mod.f(xhat,u,zeros(nw,1));
filt.xhat = xhat;
filt.P = F*filt.P*F'+G*mod.Q*G';
end