% Example:
% Makes a EKF for the following model:
% x1_t+1 = x1_t + u1_t + w1_t  
% x2_t+1 = x2_t + u2_t + w2_t  
% z_t    = sqrt(x1_t^2 + x2_t^2) + v_t
clear all
clc

% Define nonlinear model
mod.nx = 2; % Number of states
mod.nw = 2; % Dimension of process noise w
mod.f = @(x,u,w) ...
    [x(1) + u(1) + w(1); ...
     x(2) + u(2) + w(2)];
mod.h = @(x,u) ...
    [sqrt(x(1)^2 + x(2)^2)];

mod.fx = @(x,u,w) [1 0; 0 1];  % Jacobian of df/dx
mod.fw = @(x,u,w) [1 0; 0 1];  % Jacobian of df/dw

mod.hx = @(x,u) [x(1)/sqrt(x(1)^2 + x(2)^2), x(2)/sqrt(x(1)^2 + x(2)^2)]; % Jacobian of dh/dx

% Initial conditions
x0 = [1; 0];
P0 = eye(2);

% Define covariance matrices
mod.R = 1;
mod.Q = [1 .5; .5 2];

% Simulate some data

% Generate input data
u1 = [ones(1,50) 0.8*zeros(1,50) -2*ones(1,50)];
u2 = [zeros(1,50) 0.8*ones(1,50) -ones(1,50)];
u = [u1;u2];

% Generate noise inputs
% Process noise
Q = mod.Q;  
Sigma = chol(Q);
w = (randn(150,2)*Sigma)';
% Measurement noise
R = mod.R;
v = (randn(150,1)*chol(R))';

% Simulate output data
n = size(u,2);
x = zeros(mod.nx,n);
y = zeros(numel(mod.h),n);
x(:,1) = x0;
for k = 1:n-1
    y(:,k) = mod.h(x(:,k),u(:,k))+ v(:,k);
    x(:,k+1) = mod.f(x(:,k),u(:,k),zeros(2,1)) ;
end
y(:,n) = mod.h(x(:,n),u(:,n))+ v(:,n);

data.u = u;
data.y = y;

% Perform EKF
[xhat,yhat,P]=EKF(mod, data, x0, P0);

% Plot state estimate
figure(1)
plot(x(1,:),x(2,:))
hold on
plot(xhat(1,:),xhat(2,:),'r')
hold off
legend('true state', 'estimated state', 'Location','NorthWest')

% Plot model prediction
figure(2)
plot([y; yhat]')
legend('measurement', 'estimated output', 'Location','NorthWest')
