clc
close all
clear all


% Battery parameters
Q = 6; % Capacity in Ah
V_oc = 4.131; % Open circuit voltage in V
cutoff_voltage = 2; % Cutoff voltage in V

% Discharge currents for different C-rates
I_01C = 0.1 * Q; % 0.1C current
I_1C = 1.0 * Q; % 1.0C current
I_3C = 3.0 * Q; % 3C current

% Time vectors (estimate maximum discharge times based on cutoff voltage)
t_01C = linspace(0, 10, 2000); % Time for 0.1C discharge (hours)
t_1C = linspace(0, 1, 1000); % Time for 1.0C discharge (hours)
t_3C = linspace(0, 1, 1000); % Time for 3C discharge (hours)

% Shepherd equation for cell voltage
Vcell = @(I, t) V_oc - 0.00388 * I - 0.0151138 * (Q ./ (Q - I * t)) * I - 0.002 * exp((0.1 / Q) * I * t);

% Calculate Vcell for different discharge rates
V_01C = Vcell(I_01C, t_01C);
V_1C = Vcell(I_1C, t_1C);
V_3C = Vcell(I_3C, t_3C);

% Plot Vcell vs time for each discharge rate
figure;
hold on;
plot(t_01C*I_01C, V_01C, 'b', 'LineWidth', 2); % Plot for 0.1C
plot(t_1C*I_1C, V_1C, 'r', 'LineWidth', 2); % Plot for 1.0C
plot(t_3C*I_3C, V_3C, 'g', 'LineWidth', 2); % Plot for 3C

% Mark cutoff voltage line
yline(cutoff_voltage, '--k', 'LineWidth', 1.5);

xlim([0 6])
ylim([0 5])
% Add labels and legend
xlabel('Discharge Capacity (A-h)');
ylabel('Cell Voltage (V)');
title('Cell Voltage vs Time for Different Discharge Rates');
legend('0.1C', '1.0C', '3C', 'Cutoff Voltage', 'Location', 'northeast');
grid on;
hold off;


%% Energy calculation for different C-rates

Vcell_3C = @(t) 4.131 - 0.00388 * 18 - 0.0151138 * (6 ./ (6 - 18 * t)) * 18 - 0.002 * exp((0.1 / 6) * 18 * t);
Vcell_1C = @(t) 4.131 - 0.00388 * 6 - 0.0151138 * (6 ./ (6 - 6 * t)) * 6 - 0.002 * exp((0.1 / 6) * 6 * t);
Vcell_01C = @(t) 4.131 - 0.00388 * 0.6 - 0.0151138 * (6 ./ (6 - 0.6 * t)) * 0.6 - 0.002 * exp((0.1 / 6) * 0.6 * t);

% Perform numerical integration from a to b
t_cutoff_3c = 0.289; % upper limit in 3C
t_cutoff_1c = 0.9568; % upper limit in 1C
t_cutoff_01c = 9.956; % upper limit in 0.1C


E_01c = 0.6*integral(Vcell_01C, 0, t_cutoff_01c); % Numerical integration
E_1c = 6*integral(Vcell_1C, 0, t_cutoff_1c); % Numerical integration
E_3c = 18*integral(Vcell_3C, 0, t_cutoff_3c); % Numerical integration

% Display the result
disp(['The Energy in 3C is: ', num2str(E_3c)]);
disp(['The Energy in 1C is: ', num2str(E_1c)]);
disp(['The Energy in 0.1C is: ', num2str(E_01c)]);
