"""
ECMModule.py

This module contains classes and methods for simulating battery models and numerical integration.
It includes implementations of numerical integrators (Euler Forward and Runge-Kutta) and a base
class for battery cell modeling.

Classes:
--------
1. EulerForward:
    A simple Euler forward integrator class for solving ordinary differential equations (ODEs).

    Methods:
    - __init__(fun): Initializes the integrator with a function `fun`.
    - set_initial_value(y, t): Sets the initial value for the integration.
    - integrate(tvec): Performs the integration over the given time vector `tvec`.
    - successful(): Returns whether the integration was successful.

2. RungeKutta:
    A simple Runge-Kutta integrator class for solving ODEs.

    Methods:
    - __init__(fun): Initializes the integrator with a function `fun`.
    - set_initial_value(y): Sets the initial value for the integration.
    - integrate(tvec, u): Performs the integration over the given time vector `tvec` and input `u`.
    - successful(): Returns whether the integration was successful.

3. CellBase:
    An abstract base class for battery cell modeling.

    Attributes:
    - u: Input to the battery model.
    - model: The battery model function.
    - Vocv: Open-circuit voltage (OCV) lookup table.
    - Zocv: State-of-charge (SOC) lookup table.

    Methods:
    - __init__(): Initializes the CellBase class with a Runge-Kutta integrator.
    - set_attributes(opts): Sets attributes of the class based on a dictionary `opts`.
    - get_ocv(Qbatt): Calculates the open-circuit voltage (OCV) based on battery capacity `Qbatt`.
    - simulate(x0, tvec, u): Simulates the battery model using the given initial state `x0`,
      time vector `tvec`, and input `u`.
"""

# %%
# import libraries
import numpy as np

from scipy.interpolate import CubicSpline
from abc import ABC

# %%
# simulator class


class EulerForward:
    """Simple Euler forward integrator class"""

    def __init__(self, fun):
        self.f = fun

    def set_initial_value(self, y, t):
        self.y0 = y

    def integrate(self, tvec):

        self.y = np.zeros((len(tvec), len(self.y)))

        for ii, t in enumerate(tvec):
            if ii == 0:
                self.y[ii] = self.y0
                continue

            dt = t - tvec[ii - 1]

            self.y[ii] = self.y[ii - 1] + self.f(tvec[ii - 1], self.y[ii - 1]) * dt

        return self.successful()

    def successful(self):
        return True


class RungeKutta:
    """Simple Runge-Kutta integrator class"""

    def __init__(self, fun):
        self.f = fun

    def set_initial_value(self, y):
        self.y0 = y

    def successful(self):
        return True

    def integrate(self, tvec, u):

        y = np.zeros((len(self.y0), len(tvec)))

        for ii, t in enumerate(tvec):
            if ii == 0:
                y[:, ii] = self.y0
                continue

            dt = t - tvec[ii - 1]

            k1 = self.f(tvec[ii - 1], y[:, ii - 1], u[ii - 1])
            k2 = self.f(tvec[ii - 1] + dt / 2, y[:, ii - 1] + dt / 2 * k1, u[ii - 1])
            k3 = self.f(tvec[ii - 1] + dt / 2, y[:, ii - 1] + dt / 2 * k2, u[ii - 1])
            k4 = self.f(tvec[ii - 1] + dt, y[:, ii - 1] + dt * k3, u[ii - 1])

            y[:, ii] = y[:, ii - 1] + dt / 6 * (k1 + 2 * k2 + 2 * k3 + k4)

        return y


# %%
# Battery model class


class CellBase(ABC):

    u = None
    model = None
    Vocv = None
    Zocv = None

    def __init__(self):
        self.integrator = RungeKutta(lambda t, x, u: self.model(t, x, u))

    def set_attributes(self, opts: dict):
        for key, value in opts.items():
            setattr(self, key, value)

    def get_ocv(self, Qbatt):
        "This function calculates the OCV based on the battery capacity"

        OCV = np.interp(Qbatt, self.Zocv, self.Vocv)

        return OCV

    def simulate(
        self,
        x0: np.ndarray,
        tvec: np.ndarray,
        u: np.ndarray,
    ):

        self.integrator.set_initial_value(x0)
        self.y = self.integrator.integrate(tvec, u)
        self.t = tvec
        self.u = u
