"""
This script implements various tasks related to modeling and analyzing the behavior of a Li-ion battery cell using equivalent circuit models (ECMs). The tasks include determining battery capacity, terminal voltage, open-circuit voltage (OCV), internal resistance, and fitting ECMs to experimental data. The script is divided into multiple tasks, each focusing on a specific aspect of battery modeling.

Tasks Overview:
---------------
1. **Determining Q_full**:
    - Calculates the full capacity of the battery by integrating the current over time during discharge and charge steps.

2. **Determining the terminal voltages as a function of SOC**:
    - Computes the terminal voltage during charge and discharge steps and plots it against the state of charge (SOC).

3. **Determining the OCV curve**:
    - Derives the open-circuit voltage (OCV) curve by averaging the charge and discharge terminal voltages.

4. **Internal resistance of the battery as a function of SOC**:
    - Calculates the internal resistance of the battery and its average value across the SOC range.

5. **Thevinin model implementation**:
    - Implements the Thevinin equivalent circuit model for the battery and fits it to experimental data.

6. **Fitting the Thevinin model to the dynamic test data**:
    - Fits the Thevinin model to dynamic stress test (DST) data and evaluates its performance.

7. **Evaluating the Thevinin equivalent circuit model to experimental data**:
    - Tests the Thevinin model on different experimental datasets and calculates the root mean square error (RMSE).

8. **Dynamic RC model of the battery cell**:
    - Implements a dynamic RC equivalent circuit model for the battery, including charge transfer resistance and double-layer capacitance.

9. **Fitting the RC equivalent circuit model to experimental data (Training data)**:
    - Fits the RC model to training data and identifies its parameters.

10. **Evaluating the RC equivalent circuit model to experimental data (Test data)**:
     - Tests the RC model on different experimental datasets and calculates the RMSE.

11. **Saving the OCV as a CSV file**:
     - Saves the computed OCV-SOC data as a CSV file for use in subsequent analyses.

Modules and Libraries:
----------------------
- **NumPy**: For numerical computations.
- **Pandas**: For data manipulation and analysis.
- **Matplotlib**: For plotting and visualization.
- **SciPy**: For numerical integration, interpolation, and optimization.
- **ECMModule**: A custom module for equivalent circuit modeling.
- **time**: For measuring execution time.

Outputs:
--------
- Plots for SOC vs Voltage, OCV curves, internal resistance, and model evaluation.
- Printed results for battery capacity, internal resistance, and model parameters.
- Saved OCV-SOC data as a CSV file.
"""

"""
!!! IMPORTANT NOTE !!!
Only the code in the sections marked with "### Student code here ###" and "### End of student code ###" should be modified.
"""

# %%
# include the necessary modules

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from scipy.integrate import cumulative_trapezoid as cumtrapz
from scipy.interpolate import CubicSpline
from scipy.optimize import minimize

# local modules
import ECMModule as ecm
import time

# %%
# Task 1: Determining Q_full
# ----------------------------------------------

## load the data
data = pd.read_csv("../data/lowcurrent_test_data.csv")
# data.head()  # if you want to see the first few rows of the data then uncomment this line

## plotting the test data
fig, ax = plt.subplots(num=10)
ax.plot(data["Test_Time(s)"], data["Voltage(V)"], label="Voltage")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Voltage [V]")
ax.legend()

fig, ax = plt.subplots(num=11)
ax.plot(data["Test_Time(s)"], data["Current(A)"], label="Current")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Current [A]")
ax.legend()

fig, ax = plt.subplots(num=12)
ax.plot(data["Test_Time(s)"], data["Step_Index"], label="Step Index")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Step Index [-]")
ax.legend()

""" 
The full capacity of the Li-ion cell is determined by integrating the current over time. 
**Note** that there is a CC-CV charging profile in the data, which is done to get the battery to full capacity before the test. This part of the data should not be considered when calculating the total capacity of the battery.
With the help of the "Step_Index" column, we can identify the part of the data that corresponds to the CC-CV charging profile, and the part that corresponds to the test.
"""


def get_capacity(data):
    """
    Calculate the discharge and charge capacities from the given data.

    Parameters
    ----------
    data : pandas.DataFrame
        A DataFrame containing the battery test data. It must include the columns:
        - "Step_Index": The index of the step in the test procedure.
        - "Current(A)": The current in amperes.
        - "Test_Time(s)": The test time in seconds.

    Returns
    -------
    tuple of float
        A tuple containing:
        - Q_discharge: The discharge capacity in ampere-hours (Ah).
        - Q_charge: The charge capacity in ampere-hours (Ah).
    """

    charge_idx = data["Step_Index"] == 8
    discharge_idx = data["Step_Index"] == 6

    # These are the current and time data during the discharge and charge steps
    I_discharge = data["Current(A)"][discharge_idx]  # discharge current
    I_charge = data["Current(A)"][charge_idx]  # charge current
    t_discharge = data["Test_Time(s)"][discharge_idx]  # discharge time
    t_charge = data["Test_Time(s)"][charge_idx]  # charge time

    #### Student code here ####
    # HINT: Use the function np.trapezoid to calculate the integral of the current over time. For more information, see the documentation: https://numpy.org/devdocs/reference/generated/numpy.trapezoid.html

    Q_discharge = None  # Discharge capacity [Ah]
    Q_charge = None  # Charge capacity [Ah]

    #### End of student code ####

    return Q_discharge, Q_charge


print(
    f"The discharge and charge capacities are {get_capacity(data)[0]:.2f} and {get_capacity(data)[1]:.2f} Ah, respectively"
)

# CHECKPOINT
assert (np.array(get_capacity(data)) - np.array([2.08, 2.02]) < 1e-3).all()

### Student code here ###
Q_full = None  # full capacity of the battery [Ah]
### End of student code ###

print(f"The full capacity of the battery is {Q_full:.2f} Ah")

# %%
# Task 2: Determining the terminal voltages a function of SOC
# ----------------------------------------------

"""
Here is a simple way to get the OCV vs SOC curve using the following steps:
1. Get the voltage data during discharging
2. Get the voltage data during charging
3. The average of the two voltages is the OCV, here it is assumed that the internal resistance is constant.
**Note**: the number of data points for the charging and discharging steps are not the same, use interpolation to overcome this issue.
"""

#### Split into two sections, terminal voltages then OCV.


def get_terminal_voltage(data, Q_full):
    """
    Calculate the terminal voltage during charge and discharge steps.
    Parameters
    ----------
    data : pandas.DataFrame
        DataFrame containing the battery test data with columns "Step_Index", "Current(A)", "Voltage(V)", and "Test_Time(s)".
    Q_full : float
        Maximum capacity of the battery in Ah.
    Returns
    -------
    charge : numpy.ndarray
        2D array where the first row is the state of charge (SOC) vector during charge and the second row is the corresponding voltage.
    discharge : numpy.ndarray
        2D array where the first row is the state of charge (SOC) vector during discharge and the second row is the corresponding voltage.
    """

    discharge_idx = data["Step_Index"] == 6
    charge_idx = data["Step_Index"] == 8

    # These are the current, voltage, and time data during the discharge and charge steps
    Ic_discharge = data["Current(A)"][discharge_idx]
    Ic_charge = data["Current(A)"][charge_idx]
    vt_discharge = data["Voltage(V)"][discharge_idx]
    vt_charge = data["Voltage(V)"][charge_idx]
    t_discharge = data["Test_Time(s)"][discharge_idx]
    t_charge = data["Test_Time(s)"][charge_idx]

    #### Student code here ####
    # HINT: use the cumtrapz function (https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.cumulative_trapezoid.html) to integrate the current over time to get the SOC (as a vector).

    q_discharge = None  # capacity vector during discharge
    z_discharge = None  # convert to SOC

    q_charge = None  # capacity vector during charge
    z_charge = None  # convert to SOC

    #### End of student code ####

    charge = np.array([z_charge, vt_charge])  # charge capacity and voltage
    discharge = np.array([z_discharge, vt_discharge])  # discharge capacity and voltage

    return charge, discharge


charge, discharge = get_terminal_voltage(data, Q_full)

fig, ax = plt.subplots(num=20)
ax.plot(discharge[0] * 100, discharge[1], label="Discharge")
ax.plot(charge[0] * 100, charge[1], label="Charge")
ax.set_xlabel("SOC [%]")
ax.set_ylabel("Voltage [V]")
ax.legend()
ax.grid("on")


# %%
# Task 3: Determine the OCV curve
# ----------------------------------------------


def get_ocv(charge, discharge):
    """
    Calculate the Open Circuit Voltage (OCV) curve for a battery cell based on charge and discharge data.

    Parameters
    ----------
    charge : tuple of arrays
        A tuple containing two arrays:
        - The first array represents the state of charge (SOC) vector during charging.
        - The second array represents the corresponding voltage measurements during charging.
    discharge : tuple of arrays
        A tuple containing two arrays:
        - The first array represents the state of charge (SOC) vector during discharging.
        - The second array represents the corresponding voltage measurements during discharging.

    Returns
    -------
    np.ndarray
        A 2D array where the first row is the SOC vector (z_ocv) and the second row is the corresponding OCV values (v_ocv).

    Variables
    ---------
    z_discharge : np.ndarray
        State of charge (SOC) vector during discharging.
    z_charge : np.ndarray
        State of charge (SOC) vector during charging.
    vt_charge : CubicSpline
        Cubic spline interpolation function for the charge data.
    vt_discharge : CubicSpline
        Cubic spline interpolation function for the discharge data.
    z_ocv : np.ndarray
        SOC vector for the OCV curve, ranging from 0 to 1 with 101 points.
    I_c : float
        Current during the charge and discharge steps, in amperes [A].
    v_ocv_case_1 : np.ndarray
        OCV values for SOC values less than or equal to the maximum SOC during charging.
    R : np.ndarray
        Internal resistance vector for the cell, calculated for SOC values less than or equal to the maximum SOC during charging [Ohm].
    v_ocv_case_2 : np.ndarray
        OCV values for SOC values greater than the maximum SOC during charging.
    v_ocv : np.ndarray
        Concatenated array of OCV values for all SOC values.
    """

    # Extract the state of charge (SOC) vectors for discharge and charge
    z_discharge = discharge[0]
    z_charge = charge[0]

    # Creating interpolation objects for the charge and discharge data
    vt_charge = CubicSpline(
        z_charge, charge[1], extrapolate=False
    )  # charge data interpolation function
    vt_discharge = CubicSpline(
        z_discharge[::-1], discharge[1][::-1], extrapolate=False
    )  # discharge data interpolation function

    z_ocv = np.linspace(0, 1, 101)  # SOC vector for the OCV curve

    I_c = 0.1  # current during the charge and discharge steps [A]

    #### Student code here ####
    """
    Notes
    -----
    - `vt_charge` and `vt_discharge` are cubic spline interpolation functions created using the `CubicSpline` class
      from the `scipy.interpolate` module. These functions are used to interpolate the voltage values for given SOC
      values during charging and discharging, respectively.
    - `vt_charge` is initialized with the SOC and voltage data from the `charge` tuple, while `vt_discharge` is
      initialized with the reversed SOC and voltage data from the `discharge` tuple.
    - These interpolation functions can be used to estimate voltage values for any SOC value within the range of the
      provided data. For example:
        ```
        voltage_at_soc = vt_charge(soc_value)  # Interpolated voltage during charging at a specific SOC
        voltage_at_soc = vt_discharge(soc_value)  # Interpolated voltage during discharging at a specific SOC
        ```
    """
    # Case 1: z_ocv <= z_charge.max() (Using .max() to find the maximum value in z_charge)
    v_ocv_case_1 = None  # terminal voltage for SOC <= z_charge.max()

    # Case 2: z_ocv > z_charge.max()
    R = None  # internal resistance vector for the cell [Ohm]
    v_ocv_case_2 = None  # terminal voltage for SOC > z_charge.max()
    #### End of student code ####

    # dictionary to store the OCV curve
    v_ocv = np.concatenate([v_ocv_case_1, v_ocv_case_2])

    return np.array([z_ocv, v_ocv]), R


ocv, R_internal = get_ocv(charge, discharge)  # OCV and the internal resistance

fig, ax = plt.subplots(num=20)
ax.plot(discharge[0] * 100, discharge[1], label="Discharge")
ax.plot(charge[0] * 100, charge[1], label="Charge")
ax.plot(
    ocv[0] * 100,
    ocv[1],
    label="OCV",
    linestyle="--",
    color="black",
)
ax.set_xlabel("SOC [%]")
ax.set_ylabel("Voltage [V]")
ax.legend()
ax.grid("on")

# CHECKPOINT (OCV CHECK)
assert np.abs(np.interp(0.5, ocv[0] * 100, ocv[1]) - 2.94) < 0.01  # 50% SOC
assert ocv[1][0] < 2.77  # 0% SOC
assert 4.205 < ocv[1][-1] < 4.3  # 100% SOC

# %%
# Task 4: Internal resistance of the battery as a function of SOC
# ----------------------------------------------

### Student code here ###
Rth_avg = (
    None  # internal resistance of the Li-ion cell (or thevinin equivalent resistance)
)
### End of student code ###

print(f"The average internal resistance of the battery is {Rth_avg * 1e3:.2f} mOhm")

fig, ax = plt.subplots(num=30)
ax.plot(ocv[0][ocv[0] <= charge[0].max()] * 100, R_internal, label="Rth")
ax.hlines(  # plot the average internal resistance
    Rth_avg,
    ocv[0][0] * 100,
    ocv[0][-1] * 100,
    linestyle="--",
    color="black",
    label="Average Rth",
)
ax.set_xlabel("SOC [%]")
ax.set_ylabel("Internal resistance [$\\Omega$]")
ax.legend()
ax.grid("on")

# CHECKPOINT
assert np.abs(Rth_avg - 363.51e-3) < 1e-3

# %%
# Task 5: Thevinin model implementation
# ----------------------------------------------

# loading the training data
training_data = pd.read_csv("../data/training_data_DST_25C_80SOC.csv")

# plot the training data
fig, ax = plt.subplots(num=40)
ax.plot(training_data["Test_Time(s)"], training_data["Voltage(V)"], label="Voltage")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Voltage [V]")
ax.legend()

fig, ax = plt.subplots(num=41)
ax.plot(training_data["Test_Time(s)"], training_data["Current(A)"], label="Current")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Current [A]")
ax.legend()

fig, ax = plt.subplots(num=42)
ax.plot(training_data["Test_Time(s)"], training_data["Step_Index"], label="Step Index")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Step Index [-]")
ax.legend()

"""
From the training data, it is clear that after step index 6, the dynamic sterss test (DST) starts, which is the part of the data that we are interested in.
"""
meas = training_data[training_data["Step_Index"] > 6]


# Class for the thevinin equivalent circuit model of the cell
class ThevininModel(ecm.CellBase):
    """
    ThevininModel represents an equivalent circuit model (ECM) for a battery cell.

    Parameters
    ----------
    Rth : float
        Thevenin resistance of the battery cell.
    ocv : 2Darray
        Open circuit voltage (OCV) parameters, where ocv[0] is the SOC vector and ocv[1] is the OCV vector.
    Qmax : float
        Maximum capacity of the battery cell.

    Methods
    -------
    model(t, x, u)
        Computes the state of charge (SOC) change rate based on the current input.
    terminal_voltage()
        Calculates the terminal voltage of the battery cell.
    fit(data)
        Fits the model parameters to the provided data using linear regression.
    """

    def __init__(self, Rth, ocv, Q_full):
        """
        Initialize the class with Thevenin resistance, open-circuit voltage, and maximum charge capacity.

        Parameters
        ----------
        Rth : float
            Thevenin resistance.
        ocv : 2Darray
            A 2Darray containing the SOC and the corresponding open-circuit voltage vectors.
        Q_full : float
            Maximum charge capacity.
        """
        super().__init__()
        self.Rth = Rth
        self.Zocv = ocv[0]
        self.Vocv = ocv[1]
        self.Q_full = Q_full

    def model(self, t, x, u):
        """
        Calculate the change in state of charge (SOC) of a battery.

        Parameters
        ----------
        t : float
            Time variable (not used in the current implementation).
        x : float
            State variable (not used in the current implementation).
        u : float
            Input current to the battery in amperes.

        Returns
        -------
        z_dot : float
            Rate of change in state of charge (SOC) as a percentage.
        """
        Ic = u
        Q_full = self.Q_full

        #### Student code here ####
        z_dot = None
        #### End of student code ####

        return z_dot  # in % SOC

    def terminal_voltage(self):
        """
        Calculate the terminal voltage of the battery.

        Parameters
        ----------
        self : object
            The instance of the class containing battery parameters and methods.

        Returns
        -------
        float
            The terminal voltage of the battery [V].
        """

        ibatt = self.u  # battery current [A]
        zcell = self.y[0]
        vocv = self.get_ocv(zcell)  # a function of SOC
        R = self.Rth

        #### Student code here ####
        vt = None
        #### End of student code ####

        return vt

    def fit(self, data):
        """
        Fit the model to the provided data.

        Parameters
        ----------
        data : pandas.DataFrame
            The input data containing the columns "Test_Time(s)", "Current(A)", and "Voltage(V)".

        Returns
        -------
        None

        Notes
        -----
        This function uses the SLSQP method to minimize the difference between the measured voltage and the model's terminal voltage.
        The fitting process may take some time, and progress is printed to the console.
        """

        print("Model fitting started")
        print(" . . . ")
        print(" This may take a while")
        print(" . . . ")

        def funtion_to_minimize(x, data):

            tvec = data["Test_Time(s)"].values
            u = data["Current(A)"].values

            self.Rth = x[0]
            x0 = np.array([x[1]])

            self.simulate(x0, tvec, u)
            v_model = self.terminal_voltage()

            v_meas = data["Voltage(V)"].values

            return np.linalg.norm(v_model - v_meas)

        start_time = time.time()

        def scale_to_unit(x, lb, ub):
            return (x - lb) / (ub - lb)

        def unscale_from_unit(x_scaled, lb, ub):
            return x_scaled * (ub - lb) + lb

        def obj_scaled(x_scaled, data):
            x_unscaled = unscale_from_unit(x_scaled, lb, ub)
            return funtion_to_minimize(x_unscaled, data)

        lb = np.array([1e-6, 0])
        ub = np.array([200e-3, 1])
        x0 = scale_to_unit(np.array([self.Rth, 1]), lb, ub)

        res = minimize(
            lambda x: obj_scaled(x, data),
            x0,
            method="SLSQP",
            options={"disp": True},
            bounds=[
                (0, 1),
                (0, 1),
            ],
        )
        res.x = unscale_from_unit(res.x, lb, ub)

        end_time = time.time()
        print(f"Model fitting completed in {end_time - start_time:.2f} seconds")

        self.R0_avg = res.x[0]
        self.y0 = np.array([res.x[-1]])


#### testing the model
t_test_vec = np.linspace(0, 3600, 1000)
I_test = (
    np.sin(2 * np.pi * 1 / 3600 * t_test_vec) > 0
) * 2 - 1  # square wave between -5 and 5

thevinin_model = ThevininModel(Rth_avg, ocv, Q_full)
thevinin_model.simulate(np.array([0.5]), t_test_vec, I_test)
v_model = thevinin_model.terminal_voltage()

fig, ax = plt.subplots(num=50)
ax.plot(t_test_vec / 3600, v_model, label="Model")
ax.set_xlabel("Time [h]")
ax.set_ylabel("Voltage [V]")
ax.legend()

# CHECKPOINT
# if you get an error here, it is likely that the model implementation is incorrect
assert v_model[499] > 4.22
assert v_model[-1] < 3.35

# %%
# Task 6: Fitting the Thevinin model to the dynamic test data
# ----------------------------------------------
#### Create an instance of the ThevininModel class to fit the model to the training data
thevinin_model = ThevininModel(Rth_avg, ocv, Q_full)
thevinin_model.fit(meas)

# CHECKPOINT
# if you get an error here, it is likely that the model implementation is incorrect
Rth_DST = thevinin_model.Rth
assert np.abs(Rth_DST - 84.9e-3) < 10e-3

#### simulate the thevinin model
thevinin_model.simulate(
    thevinin_model.y0 * 1.01, meas["Test_Time(s)"].values, meas["Current(A)"].values
)
v_model = thevinin_model.terminal_voltage()

fig, ax = plt.subplots(num=60)
ax.plot(
    (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600,
    meas["Voltage(V)"],
    label="Measured",
)
ax.plot(
    (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600, v_model, label="Model"
)
ax.set_xlabel("Time [h]")
ax.set_ylabel("Voltage [V]")
ax.legend()

# print the value of the internal resistance of the Thevinin equivalent circuit model of the Li-ion cell
print(f"The internal resistance of the battery is {Rth_DST * 1e3:.2f} mOhm")

# %%
# Task 7: Evaluating the Thevinin equivalent circuit model to experimental data
# ----------------------------------------------

testdata_filenames = [
    "../data/test_data_FUDS_SP20_25C_50SOC.csv",
    "../data/test_data_US06_SP20_25C_80SOC.csv",
    "../data/test_data_FUDS_SP20_0C_80SOC.csv",
]
Z0 = [0.51, 0.8, 0.8]  # initial SOC
for iter, (filename, X0) in enumerate(zip(testdata_filenames, Z0)):

    data = pd.read_csv(filename)
    meas = data[data["Step_Index"] > 6]

    cell_model = ThevininModel(Rth_DST, ocv, Q_full)
    cell_model.simulate(
        np.array([X0]), meas["Test_Time(s)"].values, meas["Current(A)"].values
    )
    v_model = cell_model.terminal_voltage()

    fig, ax = plt.subplots(num=70 + iter)
    ax.plot(
        (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600,
        meas["Voltage(V)"],
        label="Measured",
    )
    ax.plot(
        (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600,
        v_model,
        label="Model",
    )
    ax.set_xlabel("Time [h]")
    ax.set_ylabel("Voltage [V]")
    ax.set_title(f"Test data {iter + 1}")
    ax.legend()

    print(
        f"RMSE for test data {iter + 1} is {np.linalg.norm(v_model - meas['Voltage(V)'].values) / np.sqrt(len(v_model)) * 1e3:0.2f} mV."
    )

# %%
# Task 8: Dynamic RC model of the battery cell
# ----------------------------------------------


class RC_model(ecm.CellBase):
    """
    A dynamic RC model of a Li-ion cell.

    Parameters
    ----------
    parameters : list or array-like
        List of parameters [R0, R1, C1] where:
        R0 : float
            Ohmic resistance [Ohms]
        R1 : float
            Charge transfer resistance [Ohms]
        C1 : float
            Double-layer capacitance [Farads]
    ocv : list or array-like
        Open circuit voltage data [Zocv, Vocv] where:
        Zocv : array-like
            State of charge values
        Vocv : array-like
            Corresponding open circuit voltage values
    Q_full : float
        Maximum cell capacity [Ah]

    Methods
    -------
    model(t, x, u)
        Computes the state derivatives for the RC model.
    terminal_voltage()
        Computes the terminal voltage of the cell.
    fit(data)
        Fits the model parameters to the provided data.

    """

    def __init__(self, parameters, ocv, Q_full):
        """
        Initialize the class with given parameters, open-circuit voltage (OCV) values, and maximum charge capacity.

        Parameters
        ----------
        parameters : list or array-like
            A list or array containing the parameters [R0, R1, C1].
        ocv : array-like
            A array containing the open-circuit voltage values [Zocv, Vocv].
        Q_full : float
            The maximum charge capacity.

        """
        super().__init__()
        self.R0 = parameters[0]
        self.R1 = parameters[1]
        self.C1 = parameters[2]
        self.Zocv = ocv[0]
        self.Vocv = ocv[1]
        self.Q_full = Q_full

    def model(self, t, x, u):
        """
        Computes the state derivatives for the given state and input.

        Parameters
        ----------
        t : float
            Current time (not used in this model but included for compatibility with ODE solvers).
        x : array_like
            State vector where:
            - x[0] is the capacitor voltage [V]
            - x[-1] is the cell capacity [Ah]
        u : float
            Input current to the cell [A]

        Returns
        -------
        ndarray
            Derivatives of the state vector:
            - vc_dot: Derivative of the capacitor voltage [V/s]
            - qbatt_dot: Derivative of the cell capacity in % SOC [1/s]
        """

        vc = x[0]  # capacitor voltage [V]
        # qbatt = x[-1] # cell capacity [Ah]
        Ic = u  # cell current [A]

        R1 = self.R1
        C1 = self.C1

        Q_full = self.Q_full

        #### Student code here ####
        z_dot = None
        vc_dot = None
        #### End of student code ####

        return np.array([vc_dot, z_dot])

    def terminal_voltage(self):
        """
        Calculate the terminal voltage of the system.

        The terminal voltage is computed using the open-circuit voltage (vocv),
        the internal resistance (R0), the current (I), and the voltage across
        the capacitor (vc).

        Returns
        -------
        float
            The terminal voltage of the system.
        """

        I = self.u
        z = self.y[-1]
        vocv = self.get_ocv(z)
        R0 = self.R0
        vc = self.y[0]

        #### Student code here ####
        v_terminal = None
        #### End of student code ####

        return v_terminal

    def fit(self, data):
        """
        Fit the model to the provided data.

        Parameters
        ----------
        data : pandas.DataFrame
            The input data containing the columns "Test_Time(s)", "Current(A)", and "Voltage(V)".

        Returns
        -------
        None

        Notes
        -----
        This function uses the SLSQP method to minimize the difference between the measured voltage and the model's terminal voltage.
        The fitting process may take some time, and progress is printed to the console.
        """

        print("Model fitting started")
        print(" . . . ")
        print(" This may take a while")
        print(" . . . ")

        def funtion_to_minimize(x, data):

            tvec = data["Test_Time(s)"].values
            u = data["Current(A)"].values

            self.R0 = x[0]
            self.R1 = x[1]
            self.C1 = x[2]
            x0 = np.array([x[3], x[-1]])

            self.simulate(x0, tvec, u)
            v_model = self.terminal_voltage()

            v_meas = data["Voltage(V)"].values

            return np.linalg.norm(v_model - v_meas)

        start_time = time.time()

        def scale_to_unit(x, lb, ub):
            return (x - lb) / (ub - lb)

        def unscale_from_unit(x_scaled, lb, ub):
            return x_scaled * (ub - lb) + lb

        def obj_scaled(x_scaled, data):
            x_unscaled = unscale_from_unit(x_scaled, lb, ub)
            return funtion_to_minimize(x_unscaled, data)

        lb = np.array([1e-6, 1e-6, 1e-6, -1, 0])
        ub = np.array([100e-3, 30e-3, 2000, 1, 1])
        x0 = scale_to_unit(np.array([self.R0, self.R1, self.C1, 0, 1]), lb, ub)

        res = minimize(
            lambda x: obj_scaled(x, data),
            x0,
            method="SLSQP",
            options={"disp": True},
            bounds=[
                (0, 1),
                (0, 1),
                (0, 1),
                (0, 1),
                (0, 1),
            ],
        )
        res.x = unscale_from_unit(res.x, lb, ub)

        end_time = time.time()
        print(f"Model fitting completed in {end_time - start_time:.2f} seconds")

        self.R0_avg = res.x[0]
        self.R1 = res.x[1]
        self.C1 = res.x[2]
        self.y0 = np.array([res.x[3], res.x[4]])


# testing the model
t_test_vec = np.linspace(0, 600, 1000)
I_test = (
    np.sin(2 * np.pi * 1 / 600 * t_test_vec) > 0
) * 2 - 1  # square wave between -5 and 5

RC_mdl = RC_model([73e-3, 20e-3, 950], ocv, Q_full)
RC_mdl.simulate(np.array([0, 0.5]), t_test_vec, I_test)
v_RC_model = RC_mdl.terminal_voltage()

# CHECKPOINT
test_R_mdl = ThevininModel(73e-3 + 20e-3, ocv, Q_full)
test_R_mdl.simulate(np.array([0.5]), t_test_vec, I_test)
v_R_mdl = test_R_mdl.terminal_voltage()

fig, ax = plt.subplots(num=80)
ax.plot(t_test_vec[1:-1], v_RC_model[1:-1], label="RC Model")
ax.plot(t_test_vec[1:-1], v_R_mdl[1:-1], label="R Model")
ax.set_xlabel("Time [s]")
ax.set_ylabel("Voltage [V]")
ax.set_title("Verification: RC Model vs R Model", fontsize=10)
ax.legend()

# %%
# Task 9: Fitting the RC equivalent circuit model to experimental data (Training data)
# ----------------------------------------------

# load the training data
"""
From the training data, it is clear that after step index 6, the dynamic sterss test (DST) starts, which is the part of the data that we are interested in.
"""
meas = training_data[training_data["Step_Index"] > 6]

# Create an instance of the RC_model class
RC_mdl = RC_model([73e-3, 20e-3, 950], ocv, Q_full)
RC_mdl.fit(meas)

# CHECKPOINT
R0_RCMdl = RC_mdl.R0
R1_RCMdl = RC_mdl.R1
C1_RCMdl = RC_mdl.C1

Y0_RCMdl = RC_mdl.y0

assert np.abs(R0_RCMdl - 75e-3) < 5e-3
assert np.abs(R1_RCMdl - 23e-3) < 1e-3
assert np.abs(C1_RCMdl - 1000) < 60

# simulate the dynamic model

RC_mdl.simulate(RC_mdl.y0, meas["Test_Time(s)"].values, meas["Current(A)"].values)
v_model = RC_mdl.terminal_voltage()

fig, ax = plt.subplots(num=90)
ax.plot(
    (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600,
    meas["Voltage(V)"],
    label="Measured",
)
ax.plot(
    (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600, v_model, label="Model"
)
ax.set_xlabel("Time [h]")
ax.set_ylabel("Voltage [V]")
ax.legend()

# print the value of the identified parameters of the RC equivalent circuit model of the Li-ion cell
print(
    f"The parameters of the RC model are R0 = {RC_mdl.R0 * 1e3:.2f} mOhm, R1 = {RC_mdl.R1 * 1e3:.2f} mOhm, and C1 = {RC_mdl.C1:.2f} F."
)

# %%
# Task 10: Evaluating the RC equivalent circuit model to experimental data (Test data)
# ----------------------------------------------

testdata_filenames = [
    "../data/test_data_FUDS_SP20_25C_50SOC.csv",
    "../data/test_data_US06_SP20_25C_80SOC.csv",
    "../data/test_data_FUDS_SP20_0C_80SOC.csv",
]
Z0 = [0.51, 0.8, 0.8]  # initial SOC

for iter, (filename, X0) in enumerate(zip(testdata_filenames, Z0)):
    data = pd.read_csv(filename)
    meas = data[data["Step_Index"] > 6]

    cell_model = RC_model([R0_RCMdl, R1_RCMdl, C1_RCMdl], ocv, Q_full)
    cell_model.simulate(
        np.array([0, X0]), meas["Test_Time(s)"].values, meas["Current(A)"].values
    )
    v_model = cell_model.terminal_voltage()

    fig, ax = plt.subplots(num=100 + iter)
    ax.plot(
        (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600,
        meas["Voltage(V)"],
        label="Measured",
    )
    ax.plot(
        (meas["Test_Time(s)"] - meas["Test_Time(s)"].iloc[0]) / 3600,
        v_model,
        label="Model",
    )
    ax.set_xlabel("Time [h]")
    ax.set_ylabel("Voltage [V]")
    ax.set_title(f"Test data {iter + 1}")
    ax.legend()

    print(
        f"RMSE test data {iter + 1} is {np.linalg.norm(v_model - meas['Voltage(V)'].values) / np.sqrt(len(v_model)) * 1e3:0.2f} mV."
    )

# %%
# show plots
plt.show()

# %%
# Saving the OCV as a csv file (used in the next lab)
# ----------------------------------------------
pd.DataFrame(
    {
        "SOC": ocv[0],
        "OCV": ocv[1],
    }
).to_csv("../data/OCV_SOC.csv", index=False)

# %%
