# Python Development Guide: Virtual Environments and Interactive Cells
This guide provides a comprehensive overview of creating and managing Python virtual environments, as well as using interactive cells in Python files with Visual Studio Code (VSCode). It is designed to help you set up a clean and efficient development environment for your Python projects.

## Table of Contents
1. [Creating and Managing Python Virtual Environments](#part-1-python-virtual-environments)
    - [Introduction](#introduction)
    - [Best Practice Recommendation](#best-practice-recommendation)
    - [Installing Python and Setting Up Virtual Environments](#installing-python-and-setting-up-virtual-environments)
      - [Installing Python](#installing-python)
        - [On Windows](#on-windows)
        - [On macOS](#on-macos)
        - [On Linux](#on-linux)
      - [Installing pip](#installing-pip)
      - [Installing virtualenv (Optional)](#installing-virtualenv-optional)
    - [Creating a Virtual Environment](#creating-a-virtual-environment)
      - [Using venv (Python 3.3+)](#using-venv-python-33)
      - [Using virtualenv (alternative method)](#using-virtualenv-alternative-method)
    - [Activating the Virtual Environment](#activating-the-virtual-environment)
      - [On Windows](#on-windows-1)
      - [On macOS/Linux](#on-macoslinux)
    - [Installing Packages from requirements.txt](#installing-packages-from-requirementstxt)
      - [Creating a requirements.txt File](#creating-a-requirementstxt-file)
      - [Installing Packages](#installing-packages)
    - [Deactivating the Virtual Environment](#deactivating-the-virtual-environment)
    - [Virtual Environment Best Practices](#virtual-environment-best-practices)
    - [Example Workflow](#example-workflow)

2. [Using Interactive Cells in Python Files with VSCode](#part-2-using-cells-in-python-files)
    - [Introduction to Interactive Cells](#introduction-to-interactive-cells)
    - [Why Use VSCode for Python Cells](#why-use-vscode-for-python-cells)
    - [Adding Cells to Python Files](#adding-cells-to-python-files)
    - [Running Cells in VSCode](#running-cells-in-vscode)
      - [Setup Requirements](#setup-requirements)
      - [Running Individual Cells](#running-individual-cells)
      - [Running All Cells](#running-all-cells)
      - [Interactive Window](#interactive-window)
      - [Debugging Cells](#debugging-cells)
    - [Configuring Python Interactive Window](#configuring-python-interactive-window)
    - [Cell Best Practices](#cell-best-practices)

**Note:** Visual studio Code (VSCode) is *recommended* as the IDE for Python development due to its great support for both virtual environments and interactive cells. However, you can also use other IDEs like PyCharm or Jupyter Notebook for similar functionalities. The instructions in this guide are focused on VSCode.

# Part 1: Python Virtual Environments

## Introduction
Virtual environments in Python provide isolated spaces where you can install packages and dependencies specific to a project without affecting the global Python installation.

## Best Practice Recommendation
**It is highly recommended that all the labs in this course uses one dedicated virtual environment.** 
This practice ensures:
- Dependency isolation between different projects
- Reproducible environments for each specific task
- Prevention of version conflicts between packages
- Easier project sharing and collaboration
- Cleaner development workflow

## Installing Python and Setting Up Virtual Environments

### Installing Python
Before creating virtual environments, ensure Python is installed on your system. Follow the steps below based on your operating system:

#### On Windows and MacOS:
1. Download the latest Python installer from the [official Python website](https://www.python.org/).
2. *Windows:* Run the installer and ensure you check the box **"Add Python to PATH"** during installation.
3. Verify the installation by opening Command Prompt and running:
    ```bash
    python --version # (or) python3 --version on macOS
    ```

#### On Linux:
1. Use your distribution's package manager to install Python. For example, on Ubuntu:
    ```bash
    sudo apt update
    sudo apt install python3-pip python3-dev python3-tk python3-venv
    ```
2. Verify the installation:
    ```bash
    python3 --version
    ```

### Installing pip
`pip` is Python's package manager and is typically included with Python installations. Verify `pip` is installed:
```bash
pip --version
```
If not installed, follow the instructions [here](https://pip.pypa.io/en/stable/installation/).

With Python and `pip` installed, you're ready to create and manage virtual environments.

## Creating a Virtual Environment

### Using venv (Python 3.3+)
Python's built-in `venv` module is the standard way to create virtual environments:

```bash
# For Windows
python -m venv myenv

# For macOS/Linux
python3 -m venv myenv
```

### Using virtualenv (alternative method)
If `venv` is not available, you can use `virtualenv`:

```bash
# Install virtualenv if not already installed
pip install virtualenv

# Create environment
virtualenv myenv
```

## Activating the Virtual Environment

### On Windows
```bash
myenv\Scripts\activate
```

### On macOS/Linux
```bash
source myenv/bin/activate
```

When activated, your command prompt will be prefixed with the environment name, e.g., `(myenv)`.

## Installing Packages from requirements.txt

### requirements.txt File
requirements.txt file exists in the root of lab directory. It lists all the packages and their versions required for the lab.

This is a sample requirements.txt file:
```
numpy==1.24.2
pandas==2.0.0
matplotlib>=3.7.1
requests
```

### Installing Packages
Once your virtual environment is activated:

```bash
pip install -r requirements.txt
```

## Deactivating the Virtual Environment
When you're done working on your lab:

```bash
deactivate
```

## Virtual Environment Best Practices
1. Always include a requirements.txt file with your projects
2. Keep your virtual environment directory out of version control (add to .gitignore)
3. Regularly update your requirements.txt as dependencies change
4. Consider using more advanced tools like Poetry or Pipenv for complex dependency management

## Example Workflow
```bash
# Create and activate environment
python -m venv lab_env
source lab_env/bin/activate  # or lab_env\Scripts\activate on Windows

# Install packages
pip install -r requirements.txt

# Work on your project...
code . # Open VSCode in the current directory

# Deactivate when finished
deactivate
```

# Part 2: Using Cells in Python Files

## Introduction to Interactive Cells
Interactive cells provide a notebook-like experience directly in Python files, combining the benefits of Jupyter notebooks with standard Python scripts.

## Why Use VSCode for Python Cells
**VSCode is highly recommended for Python development with cells** because it offers:
- Seamless integration with virtual environments
- Excellent cell execution capabilities
- Powerful debugging features
- Rich extension ecosystem
- Consistent experience across operating systems

## Adding Cells to Python Files
VSCode allows you to create Jupyter-like cells in regular Python files using special comments:

```python
# %% [Optional cell name]
# This is the first cell
print("Hello from cell 1")

# %%
# This is the second cell
import numpy as np
data = np.random.rand(10)
print(data)

# %%
# This is the third cell
import matplotlib.pyplot as plt
plt.plot(data)
plt.title("Random Data")
plt.show()
```

Each `# %%` comment marks the beginning of a new cell.

## Running Cells in VSCode

1. **Setup Requirements**:
   - Install the Python extension for VSCode
   - Select your virtual environment as the Python interpreter

2. **Running Individual Cells**:
   - Click the "Run Cell" button that appears above a cell
   - Use keyboard shortcut: `Shift+Enter`

3. **Running All Cells**:
   - Use the "Run All Cells" command from the Command Palette
   - Or use the cell toolbar at the top of the editor

4. **Interactive Window**:
   - Cell output appears in the "Python Interactive" window
   - Variables persist between cells, just like in Jupyter Notebooks

5. **Debugging Cells**:
   - Use "Debug Cell" instead of "Run Cell" for debugging

## Cell Best Practices
1. Use descriptive comments at the start of each cell to explain its purpose
2. Keep cells focused on specific tasks or concepts
3. Order cells logically for better readability
4. Consider using cell names for complex scripts (e.g., `# %% [Data Loading]`)
5. Remember to select your virtual environment as the Python interpreter for the cell execution